unit DND_Newton_Vehicle;
// =============================================================================
//   The Contence of this file are extracts from Sascha Willems Vehicle Demo
//   Copyright  2003-2004 by Sascha Willems - http://www.delphigl.de
//   visit the Delphi OpenGL Community - http://www.delphigl.com
// =============================================================================
//   Contents of this file are subject to the GNU Public License (GPL) which can
//   be obtained here : http://opensource.org/licenses/gpl-license.php
//   So only use this file if you fully unterstand that license!!!
// =============================================================================


interface

Uses Windows,OpenGL15,NewtonImport,Math3d,glPrimitives,glMatrixHelper;

Const
MultiSize = 5; // Makes the body and all assosiated things bigger by this ammount!
Torq = 250;
TireGravMag = -19.8;
VehicleMass =  100;
BreakTorq = Torq-300;

//Begin SW
Type
TNewtonVehicle = class;

  // TNewtonTire ===============================================================
  //  Baseclass for vehicle tires
  TNewtonTire = class
    ID        : Integer;
    Torque    : Single;
    Brake     : Single;
    Speed     : Single;
    GraphBody : TNewtonPrimitive;
    constructor Create(pVehicle : TNewtonVehicle; pMatrix : TMatrix4f; pWidth, pRadius : Single);
    procedure SetTirePhysics(const Vehicle : PNewtonJoint; TireID : Integer); virtual;
   end;

  // TNewtonTire ===============================================================
  //  Tire that can steer
  TNewtonSteeringTire = class(TNewtonTire)
    SteerAngle : Single;
    procedure SetTirePhysics(const Vehicle : PNewtonJoint; TireID : Integer); override;
   end;

  // TNewtonVehicle ============================================================
  TNewtonVehicle = class
    Tire      : array of TNewtonTire;
    Joint     : PNewtonJoint;
    Body      : PNewtonBody;
    GraphBody : TNewtonPrimitive;
    Matrix    : TMatrix4f;
    procedure AddTire(pMatrix : TMatrix4f; pWidth, pRadius : Single);
    procedure AddSteeringTire(pMatrix : TMatrix4f; pWidth, pRadius : Single);
    procedure SetSteeringAngle(pSteeringAngle : Single);
    procedure SetTorque(pTorque : Single);
    procedure SetBrake(pBrake : Single);
    procedure CreateCar;
    procedure Flip;
    procedure ProcessInput;
    constructor Create(pMatrix : TMatrix4f);
   end;
//end SW

procedure PSetTransform(const Body : PNewtonBody; const Matrix : PFloat); cdecl;
procedure PhysicsApplyGravityForce(const Body : PNewtonBody); cdecl;
procedure PTireUpdate(const Vehicle : PNewtonJoint); cdecl;

Function GetVehicleMatrix(Vehicle : TNewtonVehicle) : TMatrix4f;

implementation

Uses DND_Global;

constructor TNewtonTire.Create(pVehicle : TNewtonVehicle; pMatrix : TMatrix4f; pWidth, pRadius : Single);
var
 tireMass             : Single;
 tireSuspensionShock  : Single;
 tireSuspensionSpring : Single;
 tireSuspensionLength : Single;
 tirePin              : TVector3f;
begin
// Mass should be ~Mass of Vehicle/20
tireMass             := (VehicleMass/20);
// Suspension shock and spring should not be altered
tireSuspensionShock  := 1.0;
tireSuspensionSpring := 1.0;
// Length of suspension should not be too low compared to the vehicle
tireSuspensionLength := 0.25*MultiSize;
// the tire will spin around the lateral axis of the same tire space
TirePin := V3(0, 0, 1);
// Now add the tire to the vehicle joint and retrive it's internal ID (for later use)
ID        := NewtonVehicleAddTire(pVehicle.Joint, @PMatrix[0,0], @TirePin, tireMass, pWidth, pRadius,
		                              tireSuspensionShock, tireSuspensionSpring, tireSuspensionLength, self, 0);
// Set the maximum rate (angular) at which the wheel can rotate before loosing traction

//NewtonVehicleSetTireMaxLongitudinalSlideSpeed(pVehicle.Joint, ID, TireMaxLongitudinalSlideSpeed);
// Create a graphic primitive for this tire
GraphBody := PrimitiveManager.SpawnChamferCylinder(pRadius, pWidth, pMatrix);
end;

// =============================================================================
//  TNewtonTire.SetTirePhysics
// =============================================================================
//  This is used to update the tire's properties in the proper callback
// =============================================================================
procedure TNewtonTire.SetTirePhysics(const Vehicle : PNewtonJoint; TireID : Integer);
var
 Accel : Single;
 Omega : Single;
 speed : Single;
begin
// Get tire's angular velocity (angular velocity = omega)
Omega := NewtonVehicleGetTireOmega(Vehicle, TireID);
// Add some viscuos damp to the tire torque (this prevents out of control spin)
NewtonVehicleSetTireTorque(Vehicle, TireId, Torque-0.1*Omega);

speed := 1*MultiSize * Omega;
//speed := NewtonVehicleGetTireLateralSpeed(Vehicle, TireID);
NewtonVehicleSetTireMaxSideSleepSpeed (Vehicle, TireID, speed * 0.1);
// The side slip is usually propostinal to the tire longitudilnal speed, and tire load
NewtonVehicleSetTireSideSleepCoeficient (Vehicle, TireID, speed * 0.05);

// Braking
if Brake > 0 then
 begin
 // Get acceleration that's needed to stop the wheel
 Accel := NewtonVehicleTireCalculateMaxBrakeAcceleration(Vehicle, TireID);
 // Brake the wheel
 NewtonVehicleTireSetBrakeAcceleration(Vehicle, TireID, Accel, BreakTorq);

 			// set some side slipe as funtion of the linear speed
  speed := NewtonVehicleGetTireLongitudinalSpeed (Vehicle, TireID);
  NewtonVehicleSetTireMaxSideSleepSpeed (Vehicle, TireID, speed * 0.1);
 end;
// Reset brake
Brake := 0;

Self.speed := speed;
// Update graphics body
NewtonVehicleGetTireMatrix(Vehicle, ID, @GraphBody.Matrix[0,0]);
end;

// *****************************************************************************
//  TNewtonSteeringTire
// *****************************************************************************

// =============================================================================
//  TNewtonSteeringTire.SetTirePhysics
// =============================================================================
//  This is used to update the tire's properties in the proper callback.
// =============================================================================
procedure TNewtonSteeringTire.SetTirePhysics(const Vehicle : PNewtonJoint; TireID : Integer);
var
 CurrSteerAngle : Single;
begin
// Update steering angle
CurrSteerAngle := NewtonVehicleGetTireSteerAngle(Vehicle, TireId);
NewtonVehicleSetTireSteerAngle(Vehicle, TireId, CurrSteerAngle+(SteerAngle-CurrSteerAngle) * 0.25);
// Call inherited SetTirePhysics for the rest of the work
inherited;
end;

// *****************************************************************************
//  TNewtonVehicle
// *****************************************************************************

// =============================================================================
//  TNewtonVehicle.Create
// =============================================================================
constructor TNewtonVehicle.Create(pMatrix : TMatrix4f);
var
 Mass     : Single;
 Ixx      : Single;
 Iyy      : Single;
 Izz      : Single;
 Collider : PNewtonCollision;
 UpDir    : TVector3f;
 BoxSize  : TVector3f;
begin
// =============================================================================
// first we make a normal rigid body
// =============================================================================
// Create graphics primitive
GraphBody := PrimitiveManager.SpawnBox(V3(5*MultiSize, 2*MultiSize, 3*MultiSize), pMatrix);
// Get the vehicle physics dimensions
Mass      := VehicleMass/2;
BoxSize   := V3(5*MultiSize, 2*MultiSize, 3*MultiSize);
Ixx := Mass * (BoxSize.Y * BoxSize.Y + BoxSize.Z * BoxSize.Z) / 12;
Iyy := Mass * (BoxSize.X * BoxSize.X + BoxSize.Z * BoxSize.Z) / 12;
Izz := Mass * (BoxSize.X * BoxSize.X + BoxSize.Y * BoxSize.Y) / 12;
// In this demo, we use a normal box for the chassis of the car. You'd normally load a
// 3D model and use it's convex hull (NewtonCreateConvexHull) for the chassis collision
Collider := NewtonCreateBox(NewtonWorld, BoxSize.X,BoxSize.Y,BoxSize.Z, nil);
Body     := NewtonCreateBody(NewtonWorld, Collider);
// Release the Collider
NewtonReleaseCollision(NewtonWorld, Collider);
// Set user data (so that we can directly access the class in callbacks)
NewtonBodySetUserData(Body, self);
NewtonBodySetMaterialGroupID (Body,MatVehicle);

// Set transform callback
NewtonBodySetTransformCallback(Body, PSetTransform);
// Set the force and torque call back funtion
NewtonBodySetForceAndTorqueCallback(Body, PhysicsApplyGravityForce);
// Set mass matrix
NewtonBodySetMassMatrix(Body, Mass, Ixx, Iyy, Izz);
// Set the matrix for the rigid body
NewtonBodySetMatrix(Body, @pMatrix[0,0]);
// =============================================================================
//  second we need to add a vehicle joint to the body
// =============================================================================
UpDir := V3(pMatrix[0,1], pMatrix[1,1], pMatrix[2,1]);
Joint := NewtonConstraintCreateVehicle(NewtonWorld, @UpDir, Body);
// Set the callback for the tires
NewtonVehicleSetTireCallback(Joint, @PTireUpdate);
end;

// =============================================================================
//  TNewtonVehicle.CreateCar
// =============================================================================
//  Creates a normal 4-wheeled car based on this vehicle joint.
// =============================================================================
procedure TNewtonVehicle.CreateCar;
var
 TireMatrix : TMatrix4f;
begin
Matrix_SetIdentity(TireMatrix);
// First we create two static wheels (back)
Matrix_SetTransform(TireMatrix, V3( 2*MultiSize, -1*MultiSize,  2*MultiSize)); AddTire(TireMatrix, 0.8*MultiSize, 1*MultiSize);
Matrix_SetTransform(TireMatrix, V3( 2*MultiSize, -1*MultiSize, -2*MultiSize)); AddTire(TireMatrix, 0.8*MultiSize, 1*MultiSize);
// And now two steering tires (front)
Matrix_SetTransform(TireMatrix, V3(-2*MultiSize, -1*MultiSize, -2*MultiSize)); AddSteeringTire(TireMatrix, 0.8*MultiSize, 1*MultiSize);
Matrix_SetTransform(TireMatrix, V3(-2*MultiSize, -1*MultiSize,  2*MultiSize)); AddSteeringTire(TireMatrix, 0.8*MultiSize, 1*MultiSize);
// When we have attached all wheels, we MUST balance the tires
NewtonVehicleBalanceTires(Vehicle.Joint, TireGravMag);
end;

// =============================================================================
//  TNewtonVehicle.Flip
// =============================================================================
//  Flips the car back to correct orientation if it bended over
// =============================================================================
procedure TNewtonVehicle.Flip;
var
 tmpV : TVector3f;
 tmpM : TMatrix4f;
begin
tmpV := V3(0,0,0);
// Stop all car motions
NewtonVehicleReset(Vehicle.Joint);
NewtonBodySetVelocity(Vehicle.Body, @tmpV);
NewtonBodySetOmega(Vehicle.Body, @tmpV);
NewtonBodySetTorque(Vehicle.Body, @tmpV);
NewtonBodyGetMatrix(Vehicle.Body, @tmpM[0,0]);
tmpV := V3(tmpM[3,0], tmpM[3,1]+5, tmpM[3,2]);
Matrix_SetIdentity(tmpM);
Matrix_SetTransform(tmpM, tmpV);
tmpM[1,1] := 1;
NewtonBodySetMatrix(Vehicle.Body, @tmpM[0,0]);
Vehicle.SetTorque(0);
end;

// =============================================================================
//  TNewtonVehicle.SetSteeringAngle
// =============================================================================
//  Process all keys that are used to steer/accelerate the car.
// =============================================================================
procedure TNewtonVehicle.ProcessInput;
var
 Keys : array[0..255] of Boolean;
 i    : Integer;
begin
// Update keystates
for i := 0 to High(Keys) do
 Keys[i] := (Hi(GetAsyncKeyState(i)) = 128);
// Check keys
if Keys[vk_Left] then
 Vehicle.SetSteeringAngle( 30 * PI/180);
if Keys[vk_Right] then
 Vehicle.SetSteeringAngle(-30 * PI/180);
if Keys[vk_Up] then
 Vehicle.SetTorque(Torq);
if Keys[vk_Down] then
 Vehicle.SetTorque(-Torq);
if Keys[vk_Space] then
 Vehicle.SetBrake(1);
if Keys[vk_Return] then
 Vehicle.Flip;
// Reset car's steering and torque if no key is pressed
if (not Keys[vk_Left]) and (not Keys[vk_Right]) then
 SetSteeringAngle(0);
if (not Keys[vk_Up]) and (not Keys[vk_Down]) then
 SetTorque(0);
end;

// =============================================================================
//  TNewtonVehicle.SetSteeringAngle
// =============================================================================
procedure TNewtonVehicle.SetSteeringAngle(pSteeringAngle : Single);
var
 i : Integer;
begin
if Length(Tire) > 0 then
 for i := 0 to High(Tire) do
  if Tire[i] is TNewtonSteeringTire then
   (Tire[i] as TNewtonSteeringTire).SteerAngle := pSteeringAngle;
end;

// =============================================================================
//  TNewtonVehicle.SetTorque
// =============================================================================
//  Apply torque to the wheels. In this demo, the car is front-drive, so we
//  only apply torque to the two back wheels
// =============================================================================
// Stucuk: Cahnged to back wheels
procedure TNewtonVehicle.SetTorque(pTorque : Single);
var
 i : Integer;
 aTorque : Single;
begin
if Length(Tire) > 0 then
 for i := 0 to High(Tire) do
 if (Tire[i] is TNewtonSteeringTire) then
 begin
 aTorque := (Tire[i].Speed * 1.5 * 3.42 * 60 / 2);
 If aTorque < 1000 then
 aTorque := 1000
 else
 If aTorque > 6000 then
 aTorque := 6000;

 If aTorque < 4000 then
 aTorque := aTorque / 0.5
 else
 If aTorque >= 4000 then
 aTorque := aTorque / 1.5;


 If pTorque > 0 then
 Tire[i].Torque := aTorque
 else
 If pTorque < 0 then
 Tire[i].Torque := -aTorque
 else
 Tire[i].Torque := 0;
 end;
end;

// =============================================================================
//  TNewtonVehicle.SetBrake
// =============================================================================
// Stucuk: Updated, only if a tire is on the ground does it brake. stops car fliping in air
procedure TNewtonVehicle.SetBrake(pBrake : Single);
var
 i,tid : Integer;
begin
tid := -1;
if Length(Tire) > 0 then
 for i := 0 to High(Tire) do
 begin
 If tid = -1 then
 tid := NewtonVehicleGetFirstTireID(Vehicle.Joint)
 else
 tid := NewtonVehicleGetNextTireID(Vehicle.Joint,tid);
 If NewtonVehicleTireIsAirBorne(Vehicle.Joint,tid) = 0 then
  Tire[i].Brake := pBrake;
 end;
end;

// =============================================================================
//  TNewtonVehicle.AddTire
// =============================================================================
procedure TNewtonVehicle.AddTire(pMatrix : TMatrix4f; pWidth, pRadius : Single);
begin
SetLength(Tire, Length(Tire)+1);
Tire[High(Tire)] := TNewtonTire.Create(self, pMatrix, pWidth, pRadius);
end;

// =============================================================================
//  TNewtonVehicle.AddSteeringTire
// =============================================================================
procedure TNewtonVehicle.AddSteeringTire(pMatrix : TMatrix4f; pWidth, pRadius : Single);
begin
SetLength(Tire, Length(Tire)+1);
Tire[High(Tire)] := TNewtonSteeringTire.Create(self, pMatrix, pWidth, pRadius);
end;

procedure PSetTransform(const Body : PNewtonBody; const Matrix : PFloat); cdecl;
var
 TmpVehicle : TNewtonVehicle;
 TireID     : Integer;
 TmpTire    : TNewtonTire;
begin
// Get vehicle from the body's user data
TmpVehicle := TNewtonVehicle(NewtonBodyGetUserData(Body));
// Update matrix of chassis' graphical body
Move(Matrix^, TmpVehicle.GraphBody.Matrix[0,0], SizeOf(TMatrix4f));
// Now iterate through all wheels and update their matrices
TireID := NewtonVehicleGetFirstTireID(TmpVehicle.Joint);
if TireID <> 0 then
 repeat
 TmpTire := NewtonVehicleGetTireUserData(TmpVehicle.Joint, TireID);
 // Update matrix of tire's graphical body
 NewtonVehicleGetTireMatrix(TmpVehicle.Joint, TireID, @TmpTire.GraphBody.Matrix[0,0]);
// NewtonVehicleGetTireLateralSpeed()
 TireID  := NewtonVehicleGetNextTireID(TmpVehicle.Joint, TireID);
 until TireID = 0;
end;

procedure PhysicsApplyGravityForce(const Body : PNewtonBody); cdecl;
var
 mass  : Single;
 Ixx   : Single;
 Iyy   : Single;
 Izz   : Single;
 Force : TVector3f;
begin
NewtonBodyGetMassMatrix(Body, @mass, @Ixx, @Iyy, @Izz);
Force := V3(0, Mass * -19.8, 0);
NewtonBodySetForce(Body, @Force);
end;

procedure PTireUpdate(const Vehicle : PNewtonJoint); cdecl;
var
 TireID  : Integer;
 TmpTire : TNewtonTire;
begin
// Iterate through all tires and update them
TireID := NewtonVehicleGetFirstTireID(Vehicle);
if TireID <> 0 then
 repeat
 TmpTire := NewtonVehicleGetTireUserData(Vehicle, TireID);
 TmpTire.SetTirePhysics(Vehicle, TmpTire.ID);
 TireID  := NewtonVehicleGetNextTireID(Vehicle, TireID);
 until TireID = 0;
end;

Function GetVehicleMatrix(Vehicle : TNewtonVehicle) : TMatrix4f;
begin
NewtonBodyGetMatrix(Vehicle.Body, @Vehicle.Matrix);
Result := Vehicle.Matrix;
end;

end.
 