unit Q3BSP;

//------------------------------------------------------------------------
//
// Author      : Jan Horn
// Email       : jhorn@global.co.za
// Website     : http://www.sulaco.co.za
//               http://home.global.co.za/~jhorn
// Date        : 2 June 2002
// Version     : 1.0
// Description : Loads faces and textures from a Quake3 BSP file
//               Loads and renders lightmaps for the level
//
//------------------------------------------------------------------------
// Modifyed By : Stucuk
// Version     : 1.1
// Changes     : Added header detection (RTCW BSP's are now loaded)
//               Spawn Points are now loaded.
//               Entitys are loaded (Basic tags, Classname and Origin, easy to
//               expand)
// Version     : 1.2
// Changes     : Added Models
// Version     : 1.3
// Changes     : Added Water Entitys, IsInWater, ProcessEntitys
// Version     : 1.4
// Changes     : Added Newton Objects
//------------------------------------------------------------------------

interface

Uses OpenGL15, Windows, Classes, Math3d,NewtonImport;

  const FACE_POLYGON = 1;
  const MAX_TEXTURES = 1000;
  var Gamma : glFloat;

  type TBSPHeader = Record
    strID : Array[0..3] of Char;           // This should always be 'IBSP'
    Version : Integer;                     // This should be 0x2e for Quake 3 files
  end;

  type TBSPLump = Record
    Offset : Integer;                      // The offset into the file for the start of this lump
    Length : Integer;                      // The length in bytes for this lump
  end;

  type TBSPVertex = Record
    Position      : TVector3f;             // (x, y, z) position.
    TextureCoord  : TVector2f;             // (u, v) texture coordinate
    LightmapCoord : TVector2f;             // (u, v) lightmap coordinate
    Normal        : TVector3f;             // (x, y, z) normal vector
    Color         : Array[0..3] of Byte    // RGBA color for the vertex
  end;

  type TBSPFace = Record
    textureID : Integer;                   // The index into the texture array
    effect    : Integer;                   // The index for the effects (or -1 = n/a)
    FaceType  : Integer;                   // 1=polygon, 2=patch, 3=mesh, 4=billboard
    startVertIndex : Integer;              // The starting index into this face's first vertex
    numOfVerts     : Integer;              // The number of vertices for this face
    meshVertIndex  : Integer;              // The index into the first meshvertex
    numMeshVerts   : Integer;              // The number of mesh vertices
    lightmapID     : Integer;              // The texture index for the lightmap
    lMapCorner : Array[0..1] of Integer;   // The face's lightmap corner in the image
    lMapSize   : Array[0..1] of Integer;   // The size of the lightmap section
    lMapPos  : TVector3f;                  // The 3D origin of lightmap.
    lMapVecs : Array[0..1] of TVector3f;   // The 3D space for s and t unit vectors.
    vNormal  : TVector3f;                  // The face normal.
    Size : Array[0..1] of Integer;         // The bezier patch dimensions. 
  end;

  type TBSPTexture = Record
    TextureName : Array[0..63] of Char;    // The name of the texture w/o the extension
    flags    : Integer;                    // The surface flags (unknown)
    contents : Integer;                    // The content flags (unknown)
  end;

  type TBSPLightmap = Record
    imageBits : Array[0..127, 0..127, 0..2] of Byte;   // The RGB data in a 128x128 image
  end;

  type TBSPPosition = record
  Position : TVector3f;
  Angle : single;
  end;

  type TBSPEntity = record
  Classname,model  : String;
  Origin,Max,Min   : TVector3f;
  Mass             : integer;
  Angles           : TVector3f;
  end;

  type TBSPModel = record
  min              : TVector3f;       // The min position for the bounding box
  max              : TVector3f;       // The max position for the bounding box.
  faceIndex        : integer;         // The first face index in the model
  numOfFaces       : integer;         // The number of faces in the model
  brushIndex       : integer;         // The first brush index in the model
  numOfBrushes     : integer;         // The number brushes for the model
  end;

  PNewtonObject = ^TNewtonObject;
  TNewtonObject =  class(TObject)
  model            : integer;         // The id of the model
  Origin           : TVector3f;
  Mass             : integer;
  Matrix           : TGLMatrixf4;
  Body             : PNewtonBody;
  Index            : Integer;
  end;

  type TQuake3BSP = Object
    private

    public
      numOfVerts        : Integer;             // The number of verts in the model
      numOfFaces        : Integer;             // The number of faces in the model
      numOfTextures     : Integer;             // The number of texture maps
      numOfLightmaps    : Integer;             // The number of light maps
      numOfBSPEntitys   : Integer;             // The number of entitys
      numOfMeshVerts    : Integer;             // The number of meshverts
      numOfModels       : Integer;             // The number of Models
      numOfWaterEntitys : Integer;             // The number of WaterEntitys
      numOfNewtonObjects: Integer;             // The number of NewtonObjects
      Vertices          : Array of TBSPVertex; // The object's vertices
      MeshVertices      : Array of Integer;
      Faces             : Array of TBSPFace;   // The faces information of the object
      Models            : array of TBSPModel;

      EntityList        : TStringList;            // All the Entities of the level
      Entitys           : array of TBSPEntity;
      WaterEntitys      : Array of Integer;
      NewtonObjects     : Array of TNewtonObject;

      IBSP : Boolean;        // So we know what type it is, IBSP or RBSP

      // Spawn Points
      numOfPositions : integer;
      Positions : array of TBSPPosition;
      curPosition : integer; // Current SpawnPoint

      //Texture Stuff
      LastTex : Integer;

      // The texture and lightmap array for the level
      Textures  : Array of glUint;
      procedure RenderFace(faceIndex : Integer);
      function  LoadBSP(const FileName : String) : Boolean;
      procedure RenderLevel(const Pos : TVector3f);
      procedure LoadEntities(var f:file;Lump:TBSPLump);
      procedure SetPositions(l : TStringList);
      function  PlayerPosition : TBSPPosition;
      procedure ProcessEntity(Block : TStringList);
      Function  ParseEntityVector3f(const S : String) : TVector3f;
      Function  ParseEntityStringValue(const S : String; var Rest : string) : string;
      Function  ReturnOrigin(Classname : String) : TVector3f;
      procedure SwapBoundingBoxFloat(var min, max : TVector3f);
      Function  GetWaterModel : Integer;
      Procedure ProcessEntitys;
      Function  IsInWater(Position : TVector3f) : Boolean;
      Function  StripOriginFromModel(ModelId : Integer) : TVector3f;
      procedure RenderFaceModel(faceIndex : Integer; Pos,COB : TVector3f);
      procedure RenderEntityModel(const Pos: TVector3f; ModelId : Integer);
      Procedure RenderWater;
  end;

  // This is our lumps enumeration
  const
    kEntities    = 0;            // Stores player/object positions, etc...
    kTextures    = 1;            // Stores texture information
    kPlanes      = 2;            // Stores the splitting planes
    kNodes       = 3;            // Stores the BSP nodes
    kLeafs       = 4;            // Stores the leafs of the nodes
    kLeafFaces   = 5;            // Stores the leaf's indices into the faces
    kLeafBrushes = 6;            // Stores the leaf's indices into the brushes
    kModels      = 7;            // Stores the info of world models
    kBrushes     = 8;            // Stores the brushes info (for collision)
    kBrushSides  = 9;            // Stores the brush surfaces info
    kVertices    = 10;           // Stores the level vertices
    kMeshVerts   = 11;           // Stores the model vertices offsets
    kShaders     = 12;           // Stores the shader files (blending, anims..)
    kFaces       = 13;           // Stores the faces for the level
    kLightmaps   = 14;           // Stores the lightmaps for the level
    kLightVolumes= 15;           // Stores extra world lighting information
    kVisData     = 16;           // Stores PVS and cluster info (visibility)
    kMaxLumps    = 17;           // A constant to store the number of lumps

    Var
    inputBuffer       : Array of TBSPVertex;

implementation

Uses SysUtils, Textures, DND_Global, DND_Newton;

{ TQuake3BSP }

Function TQuake3BSP.ParseEntityStringValue(const S : String; var Rest : string) : string;
var
i,st,e,SLength : integer;
begin

st := -1;
e := -1;
SLength := length(s);

for i := 1 to length(s) do
if st = -1 then
if copy(s,i,1) <> ' ' then
begin
st := i;
break;
end;

for i := st to length(s) do
if e = -1 then
if copy(s,i,1) = ' ' then
begin
e := i;
break;
end;

if e = -1 then
e := length(s)+1;

if st = -1 then
begin
Result := '';
Rest := '';
exit;
end;

Result := copy(s,st,e-st);
Rest := '';

if e < length(s) then
Rest := copy(s,e,length(s));
end;

Function TQuake3BSP.ParseEntityVector3f(const S : String) : TVector3f;
var
Rest,Rest2 : string;
begin

Result.X := StrToFloat(ParseEntityStringValue(S,Rest));
Result.Z := -StrToFloat(ParseEntityStringValue(Rest,Rest2));
Result.Y := StrToFloat(ParseEntityStringValue(Rest2,Rest));

end;

procedure TQuake3BSP.ProcessEntity(Block : TStringList);
var
i,s,e,s2,e2 : integer;
Name,Value,ValueUC,SS : string;
Angle : Boolean;
begin

if not (block.count > 0) then exit;

inc(numOfBSPEntitys);
Setlength(Entitys,numOfBSPEntitys);

Angle := False;

// Process data and place in Entitys

try
for i := 0 to block.Count-1 do
begin

s := Pos('"',block.Strings[i]);
e := Pos('"',copy(block.Strings[i],s+1,length(block.Strings[i])));

Name := ansilowercase(copy(block.Strings[i],s+1,e-s));

SS := copy(block.Strings[i],e+2,length(block.Strings[i]));

s2 := Pos('"',SS);

e2 := Pos('"',copy(SS,s2+1,length(SS)));

ValueUC := copy(SS,s2+1,e2+1-s2);
Value := ansilowercase(ValueUC);


if name = 'classname' then
Entitys[numOfBSPEntitys-1].Classname := Value
else
if name = 'model' then
Entitys[numOfBSPEntitys-1].model := Value
else
if name = 'origin' then
Entitys[numOfBSPEntitys-1].Origin := ParseEntityVector3f(Value)
else
if name = 'mass' then
Entitys[numOfBSPEntitys-1].mass := StrtoInt(Value)
else
begin
if (name <> '') and (value <> '') then
begin
// Need to add support for unknowns sometime (used in func_multipul)
{inc(Entitys[numOfBSPEntitys-1].Unknown_No);
SetLength(Entitys[numOfBSPEntitys-1].Unknowns,Entitys[numOfBSPEntitys-1].Unknown_No);
Entitys[numOfBSPEntitys-1].Unknowns[Entitys[numOfBSPEntitys-1].Unknown_No-1].Name := Name;
Entitys[numOfBSPEntitys-1].Unknowns[Entitys[numOfBSPEntitys-1].Unknown_No-1].Value := Value;  }
end;
end;
end
except
// Add Error Message Here!
end;

end;

function StrToVector3f(s : string) : TVector3f;
var v : TVector3f;
    p,l : integer;
    tmp : string;
begin
  l := Length(s);
  if l > 0 then begin
    p := 1;
    tmp := '';
    while s[p] <> ' ' do begin
      tmp := tmp + s[p];
      inc(p);
    end;
    v.x := StrToInt(tmp);
    inc(p);

    tmp := '';
    while s[p] <> ' ' do begin
      tmp := tmp + s[p];
      inc(p);
    end;
    v.z := -StrToInt(tmp);
    inc(p);
    tmp := '';
    while p <= l do begin
      tmp := tmp + s[p];
      inc(p);
    end;
    v.y := StrToInt(tmp);
  end;
  result := v;
end;

function TQuake3BSP.PlayerPosition : TBSPPosition;
begin
  if numOfPositions > 0 then begin
    result := Positions[curPosition];
    inc(curPosition);
    if curPosition >= numofPositions then
      curPosition := 0;
  end;
end;

procedure TQuake3BSP.SetPositions(l : TStringList);
var i : integer;
    v, p : string;
begin
  if l.Count > 0 then begin
    inc(numOfPositions);
    SetLength(Positions, numOfPositions);
    i := 0;
    while i < l.Count do begin
      v := l[i];
      p := l[i+1];
      if v = 'origin' then
        Positions[numOfPositions-1].Position := StrToVector3f(p)
      else if v = 'angle' then
        Positions[numOfPositions-1].Angle := StrToFloat(p)
      else if v = 'spawnflags' then
        curPosition := numOfPositions-1;
      inc(i, 2);
    end;
  end;
end;

procedure TQuake3BSP.LoadEntities(var f:file;Lump:TBSPLump);
var ents : array of char;
    tmp, value, name, entityName, class_name : string;
    i, cnt, k, idx : integer;
    block : TStringList;
    s,e,ii : integer;
begin
  numOfBSPEntitys := 0;
  SetLength(ents, lump.Length);
  EntityList.Clear;
  block := TStringList.Create;
  Seek(f,lump.offset);
  BlockRead(f,ents[0],lump.Length);

  curPosition := 0;

  // scan all chars
  cnt := 0;
  for i := 0 to lump.Length-1 do begin
    case ents[i] of
      '{', #10, #13, #9 : ; // NOP
      '}' : begin
              idx := block.IndexOf('classname')+1;
              class_name := block[idx];
              if (class_name = 'info_player_deathmatch') or (class_name = 'info_player_start') then
                SetPositions(block);
              block.Clear;
            end;
      '"' : begin
              inc(cnt);
              if cnt = 1 then begin
                tmp := '';
              end
              else begin
                cnt := 0;
                if not Odd(block.Count) then
                  block.Add(lowerCase(tmp))
                else
                  block.Add(tmp);

                tmp := '';
              end;
            end;
      else begin tmp := tmp + ents[i]; end;
    end;
  end;

  tmp := '';

  for i := 0 to Lump.Length-1 do
  tmp := tmp + ents[i];


  EntityList.Text := tmp;
  //EntityList.SaveToFile(GAME_FOLDER + '\entitys.txt');

  s := -1;
  e := -1;

  if EntityList.Count > 0 then
  for i := 0 to EntityList.Count-1 do
  begin

  if (s = -1) and (Pos('{',EntityList.Strings[i]) <> 0) then
  s := i;

  if (e = -1) and (Pos('}',EntityList.Strings[i]) <> 0) then
  e := i;

  if (s > e) and (e > -1) then
  begin
  s := -1;
  e := -1;
  end;

  if (s <> -1) and (e <> -1) then
  begin
  block.Clear;
  for k := s to e do
  block.Add(EntityList.Strings[k]);

  s := -1;
  e := -1;

  if block.Count > 0 then
  ProcessEntity(Block);
  end;

  end;

  block.Free;

end;

{-----------------------------------------------------------}
{---  This loads in all of the .bsp data for the level   ---}
{-----------------------------------------------------------}
function TQuake3BSP.LoadBSP(const FileName: String): Boolean;
var F : File;
    I : Integer;
    Temp : glFloat;
    Header : TBSPHeader;
    Lumps  : Array of TBSPLump;
    BSPTextures  : Array of TBSPTexture;
begin
  result :=FALSE;

  // Check if the .bsp file can be opened
  AssignFile(F, filename);
{$I-}
  Reset(F,1);
{$I+}
  if IOResult <> 0 then
  begin
    MessageBox(0, 'Could not find the BSP file!', 'Error', MB_OK);
    exit;
  end;

  SetLength(Lumps, kMaxLumps);

  // Read in the header and lump data
  BlockRead(F, Header, Sizeof(Header));

  if (Header.strID[0] = 'I') and (Header.strID[1] = 'B') and (Header.strID[2] = 'S') and (Header.strID[3] = 'P') and ((Header.Version = 46) or (Header.Version = 47)) then
  IBSP := True
  else
  if (Header.strID[0] = 'R') and (Header.strID[1] = 'B') and (Header.strID[2] = 'S') and (Header.strID[3] = 'P') and (Header.Version = 1) then
  IBSP := False
  else
  begin
    MessageBox(0, pchar('Wrong header information! Tag: ' + Header.strID[0]+Header.strID[1]+Header.strID[2]+Header.strID[3] + ' Version: ' + inttostr(Header.Version)), 'Error', MB_OK);
    exit;
  end;

  BlockRead(F, Lumps[0], kMaxLumps*sizeof(TBSPLump));

  EntityList := TStringList.Create;

   // LUMP 0: ENTITIES **********************************************************
  try
    LoadEntities(F, lumps[kEntities]);       // Load the entities
  except
  //  ShowMessage('Error in entities');
    exit;
  end;   

  // Allocate the vertex memory
  numOfVerts := Round(lumps[kVertices].length / sizeof(TBSPVertex));
  SetLength(Vertices, numOfVerts);

  // Allocate the face memory
  numOfFaces := Round(lumps[kFaces].length / sizeof(TBSPFace));
  SetLength(Faces, numOfFaces);

  // Mesh Verts
  numOfMeshVerts := Round(lumps[kMeshVerts].length / sizeof(integer));

  numOfModels := Round(lumps[kModels].length / sizeof(TBSPModel));

  // Allocate memory to read in the texture information.
  numOfTextures := Round(lumps[kTextures].length / sizeof(TBSPTexture));
  SetLength(BSPTextures, numOfTextures);
  SetLength(Textures, numOfTextures);

  // Seek to the position in the file that stores the vertex information
  Seek(F, lumps[kVertices].offset);

  // Go through all of the vertices that need to be read and swap axises
  for I :=0 to numOfVerts-1 do
  begin
    // Read in the current vertex
    BlockRead(F, Vertices[i], sizeOf(TBSPVertex));

    // Swap the y and z values, and negate the new z so Y is up.
    Temp := Vertices[i].Position.Y;
    Vertices[i].Position.Y := Vertices[i].Position.Z;
    Vertices[i].Position.Z := -temp;

    // Negate the V texture coordinate because it is upside down otherwise...
    // Stucuk: Should be 1- to convert it from DX coords to OpenGL not just -
    Vertices[i].TextureCoord.Y := 1-Vertices[i].TextureCoord.Y;
  end;


  numOfNewtonObjects := 0;
  SetLength(NewtonObjects,numOfNewtonObjects);

  // Mesh Verts
  SetLength(MeshVertices, numOfMeshVerts);
  Seek(F, lumps[kMeshVerts].offset);
  BlockRead(F, MeshVertices[0], numOfMeshVerts*sizeof(integer));

  SetLength(inputBuffer,numOfVerts);

  // Seek and read in all the face information
  Seek(F, lumps[kFaces].offset);
  BlockRead(F, Faces[0], numOfFaces*sizeOf(TBSPFace));

  // Seek and read in all the texture information
  Seek(F, lumps[kTextures].offset);
  BlockRead(F, BSPTextures[0], numOfTextures*sizeOf(TBSPTexture));

  // LUMP 7: MODELS ************************************************************
  try
    //numOfModels := Round(lumps[kModels].length / sizeof(TBSPModel));
    SetLength(Models, numOfModels);
    Seek(F, lumps[kModels].offset);
    BlockRead(F, Models[0], sizeOf(TBSPModel)*numOfModels);
    for i := 0 to numOfModels-1 do begin
      SwapBoundingBoxFloat(Models[i].min, Models[i].max);
    end;

  except
//    ShowMessage('Error in models');
    exit;
  end;

  // Go through all of the textures
  for I :=0 to numOfTextures-1 do
  begin
    // for some reason known only to ID they dont store file extentions
    // Stucuk: They refer to shader names not filenames, tho for most shaders
    //         the first texture is the shaders name
    if FileExists(BSPTextures[i].TextureName + '.jpg') then
      LoadTexture(BSPTextures[i].TextureName + '.jpg', Textures[i],False,False,False)
    else if FileExists(BSPTextures[i].TextureName + '.tga') then
      LoadTexture(BSPTextures[i].TextureName + '.tga', Textures[i],False,False,False)
  end;
  CloseFile(F);

  numOfWaterEntitys := 0;
  SetLength(WaterEntitys,0);
  ProcessEntitys;

  result :=TRUE;
end;


{-----------------------------------------------------------}
{--- Renders a face, determined by the passed in index   ---}
{-----------------------------------------------------------}
procedure TQuake3BSP.RenderFace(faceIndex: Integer);
var Face : TBSPFace; i : Integer;
begin
  // Here we grab the face from the index passed in
  Face := Faces[faceIndex];

  glEnable(GL_TEXTURE_2D);                // Turn on texture mapping and bind the face's texture map
  If LastTex <> Face.textureID then // Stucuk: Reduces texture binds
  begin
  glBindTexture(GL_TEXTURE_2D, textures[Face.textureID]);
  LastTex := Face.textureID;
  end;

  for i := 0 to face.numMeshVerts-1 do
  inputBuffer[i] := Vertices[Face.startVertIndex+MeshVertices[face.meshVertIndex+i]];

  // Give OpenGL our vertices to use for vertex arrays
  glVertexPointer(3, GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].Position);

  // Since we are using vertex arrays, we need to tell OpenGL which texture
  // coordinates to use for each texture pass.  We switch our current texture
  // to the first one, then set our texture coordinates.
  glTexCoordPointer(2, GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].TextureCoord);
  glNormalPointer(GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].Normal);

  // Draw the face in a triangle face, starting from the starting index
  // to the starting index + the number of vertices.  This is a vertex array function.
  glDrawArrays(GL_TRIANGLES, 0, face.numMeshVerts);
end;


{---------------------------------------------------------------------}
{--- Goes through all faces and draws them if type is FACE_POLYGON ---}
{---------------------------------------------------------------------}
procedure TQuake3BSP.RenderLevel(const Pos: TVector3f);
var i : Integer;
begin
  LastTex := -1;
  // Set our vertex array client states for vertices and texture coordinates
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  for i := Models[0].faceIndex to Models[0].faceIndex + Models[0].numOfFaces do
  if Faces[i].Facetype = FACE_POLYGON then
  RenderFace(i);

  for i := 0 to numOfNewtonObjects-1 do
  begin
  glLoadIdentity;
  Camera.ApplyCamera;
//  NewtonBodyGetMatrix(NewtonObjects[numOfNewtonObjects-1].Body,@NewtonObjects[numOfNewtonObjects-1].Matrix);
  glMultMatrixf(@NewtonObjects[i].Matrix);
//  NewtonObjects[i].Matrix[3,0]
  RenderEntityModel(SetVector(0,0,0),NewtonObjects[i].model);
  end;

  glLoadIdentity;
  Camera.ApplyCamera;

  RenderWater;
end;

procedure TQuake3BSP.RenderFaceModel(faceIndex: Integer; Pos,COB : TVector3f);
var Face : TBSPFace; i : Integer;
begin
  // Here we grab the face from the index passed in
  Face := Faces[faceIndex];

  glEnable(GL_TEXTURE_2D);                // Turn on texture mapping and bind the face's texture map
  If LastTex <> Face.textureID then // Stucuk: Reduces texture binds
  begin
  glBindTexture(GL_TEXTURE_2D, textures[Face.textureID]);
  LastTex := Face.textureID;
  end;

  for i := 0 to face.numMeshVerts-1 do
  begin
  inputBuffer[i] := Vertices[Face.startVertIndex+MeshVertices[face.meshVertIndex+i]];
  inputBuffer[i].Position := AddVector(SubtractVector(inputBuffer[i].Position,COB),Pos);
  end;

  // Give OpenGL our vertices to use for vertex arrays
  glVertexPointer(3, GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].Position);

  // Since we are using vertex arrays, we need to tell OpenGL which texture
  // coordinates to use for each texture pass.  We switch our current texture
  // to the first one, then set our texture coordinates.
  glTexCoordPointer(2, GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].TextureCoord);
  glNormalPointer(GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].Normal);

  // Draw the face in a triangle face, starting from the starting index
  // to the starting index + the number of vertices.  This is a vertex array function.
  glDrawArrays(GL_TRIANGLES, 0, face.numMeshVerts);
end;


{---------------------------------------------------------------------}
{--- Goes through all faces and draws them if type is FACE_POLYGON ---}
{---------------------------------------------------------------------}
procedure TQuake3BSP.RenderEntityModel(const Pos: TVector3f; ModelId : Integer);
var f : Integer;
begin
  // Set our vertex array client states for vertices and texture coordinates
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);

  If (ModelId > 0) and (ModelID < numOfModels) then
  for f := Models[ModelId].faceIndex to Models[ModelId].faceIndex + Models[ModelId].numOfFaces-1 do
  if Faces[f].Facetype = FACE_POLYGON then
  RenderFaceModel(f,Pos,GetCenterOfBounds(Models[ModelId].max,Models[ModelId].min));
end;

Function TQuake3BSP.ReturnOrigin(Classname : String) : TVector3f;
var
x : integer;
begin

if numOfBSPEntitys > 0 then
for x := 0 to numOfBSPEntitys-1 do
if ansilowercase(Entitys[x].Classname) = ansilowercase(Classname) then
Result := Entitys[x].Origin;

end;

procedure TQuake3BSP.SwapBoundingBoxFloat(var min, max : TVector3f);
var temp : single;
begin
  // Swap the min y and z values, then negate the new Z
  temp := min.y;
  min.y := min.z;
  min.z := -temp;
  // Swap the max y and z values, then negate the new Z
  temp := max.y;
  max.y := max.z;
  max.z := -temp;
  // swap min/max if necessary
  if min.Y > max.Y then begin
    temp := min.Y;
    min.Y := max.Y;
    max.Y := temp;
  end;
  if min.z > max.z then begin
    temp := min.z;
    min.z := max.z;
    max.z := temp;
  end;
end;

Function TQuake3BSP.GetWaterModel : Integer;
var
x : integer;
begin

for x := 0 to numOfBSPEntitys-1 do
if ansilowercase(Entitys[x].Classname) = 'func_water' then
If copy(Entitys[x].model,1,1) = '*' then
begin
Result := strtoint(copy(Entitys[x].model,2,length(Entitys[x].model)));
exit;
end;

end;

Procedure TQuake3BSP.ProcessEntitys;
var
x,mid : integer;
begin

for x := 0 to numOfBSPEntitys-1 do
begin
if ansilowercase(Entitys[x].Classname) = 'func_water' then
If copy(Entitys[x].model,1,1) = '*' then
begin
Inc(numOfWaterEntitys);
SetLength(WaterEntitys,numOfWaterEntitys);
WaterEntitys[numOfWaterEntitys-1] := x;

mid := strtoint(copy(Entitys[x].model,2,length(Entitys[x].model)));
Entitys[x].max := Models[mid].max;
Entitys[x].min := Models[mid].min;
Entitys[x].model := inttostr(mid);
end;

if ansilowercase(Entitys[x].Classname) = 'newton_object' then
If copy(Entitys[x].model,1,1) = '*' then
begin
mid := strtoint(copy(Entitys[x].model,2,length(Entitys[x].model)));

If (mid < 0) or (mid > numOfModels-1) then continue;

Inc(numOfNewtonObjects);
SetLength(NewtonObjects,numOfNewtonObjects);

NewtonObjects[numOfNewtonObjects-1] := TNewtonObject.Create;
NewtonObjects[numOfNewtonObjects-1].Index       := numOfNewtonObjects-1;
NewtonObjects[numOfNewtonObjects-1].Mass        := Entitys[x].Mass;
NewtonObjects[numOfNewtonObjects-1].model       := mid;
NewtonObjects[numOfNewtonObjects-1].Origin      := StripOriginFromModel(mid);
CreateNewtonObject(Self,mid,numOfNewtonObjects-1,Entitys[x].Angles);
end;


end;

end;

Function TQuake3BSP.IsInWater(Position : TVector3f) : Boolean;
var
x : integer;
begin
Result := False;

For x := 0 to numOfWaterEntitys-1 do
if (Position.X >= Entitys[WaterEntitys[x]].min.X) and (Position.X <= Entitys[WaterEntitys[x]].max.X) and (Position.Y >= Entitys[WaterEntitys[x]].min.Y) and (Position.Y <= Entitys[WaterEntitys[x]].max.Y) and (Position.Z >= Entitys[WaterEntitys[x]].min.Z) and (Position.Z <= Entitys[WaterEntitys[x]].max.Z) then
begin
Result := True;
Exit;
end;

end;

Function TQuake3BSP.StripOriginFromModel(ModelId : Integer) : TVector3f;
var
Origin : TVector3f;
f,v : integer;
begin

Origin := GetCenterOfBounds(Models[ModelId].max,Models[ModelId].min);

for f := Models[ModelId].faceIndex to Models[ModelId].faceIndex + Models[ModelId].numOfFaces-1 do
for v := 0 to Faces[f].numOfVerts-1 do
Vertices[Faces[f].startVertIndex+v].Position := SubtractVector(Vertices[Faces[f].startVertIndex+v].Position,Origin);

Models[ModelId].max := SubtractVector(Models[ModelId].max,Origin);
Models[ModelId].min := SubtractVector(Models[ModelId].min,Origin);

Result := Origin;
end;

Procedure TQuake3BSP.RenderWater;
var
x,y,mid : integer;
begin

glEnable(GL_BLEND);
glColor4f(1,1,1,0.5);


for x := 0 to numOfWaterEntitys-1 do
begin
mid := strtoint(Entitys[WaterEntitys[x]].model);
for y := 0 to Models[mid].numOfFaces-1 do
RenderFace(Models[mid].faceIndex + y);
end;

glColor4f(1,1,1,1);
glDisable(GL_BLEND);
end;

end.
